/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE main-dp.c - *SAT 1.1 */
/*  Main program */

/*  *********************************************************************** */
/*  *********************************************************************** */

/* Standard modules */
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

/* *SAT modules */
#include "fnode.h"
#include "dag.h"
#include "dp_sat.h"
#include "satotrie.h" 
#include "monitor.h"
#include "table.h"
#include "set.h"

/* Choose the logic consistency test package */
#ifdef TRANSLATE_TO_K
#include "kdp.h"
#else

#ifdef KSAT
#include "kdp.h"
#endif

#ifdef ESAT
#include "edp.h"
#endif

#endif

/* Main variables */
extern fnode_t  *formula_as_tree;
dag_t           *dag_ptr;
image_t         *main_image_ptr;
monitor_t       *main_monitor_ptr;
int             res;
Cons_func_ptr_t Test_star_consistency;

/* Prototypes */
extern int      yyparse();
Cons_func_ptr_t Choose_test(); 
void            Translate_to_K(fnode_t *formula_fnode_ptr);
void            Set_preprocessing_flags(int *bary, int *nnf_conversion, int *negate);
void            Setup_memoizing(dag_t *dag_ptr);
int             Skip_params(int argc, char **argv); 
void            Help_and_exit(char **argv);
void            Display_results(monitor_t *main_monitor_ptr);

int main(int argc, char **argv)
{
  int filec;
  int i;

  int bary, nnf_conversion, negate;

  /* Initializing the monitor */
  main_monitor_ptr = Init_monitor();

  /* If no argument is supplied give help and exit starSAT */
  if (argc < 2) Help_and_exit(argv);

  /* Getting the index for the file name in command line */
  filec = Skip_params(argc, argv);

  /* Parsing the input formula */
  if (freopen(argv[filec], "r", stdin) != NULL) { 
    Start_clock(main_monitor_ptr, PARSE);
    yyparse();
    Stop_clock(main_monitor_ptr, PARSE);
    /* Recover stdin!!! */
    freopen("/dev/tty", "r", stdin); 
  } else { 
    printf("File %s does not exists\n",argv[1]); 
    exit(-1); 
  } 

  /* Preprocessing */
  Start_clock(main_monitor_ptr, PREPROCESS);
  Set_preprocessing_flags(&bary, &nnf_conversion, &negate);

  if (bary) {
    if (nnf_conversion) {
      if (negate)
	Convert_NNF(formula_as_tree, 1);
      else
	Convert_NNF(formula_as_tree, 0);
    }
    /* If the input syntax defines binary trees,
       they are converted to n-ary trees */
    formula_as_tree = Convert_binary2nary(formula_as_tree);
    Flatten_and_or(formula_as_tree);  
  }
  /* Always get rid of diamonds from the input syntax */
  Convert_dia2box(formula_as_tree);
  Stop_clock(main_monitor_ptr, PREPROCESS);

#ifdef TRANSLATE_TO_K
  /* Applying a translational approach for a classical modal logic */
  Start_clock(main_monitor_ptr, TRANSLATE);
  Translate_to_K(formula_as_tree);
  Stop_clock(main_monitor_ptr, TRANSLATE);
#endif

  Start_clock(main_monitor_ptr, BUILD);
  if (!bary) {
    if (negate)
      /* Working as a validity checker. The input formula is negated. */
      dag_ptr = Make_dag(Make_formula_nary(not_code, empty_code, formula_as_tree), 0, NCDE);  
    else
      /* Working as a satisfiability checker. */
      dag_ptr = Make_dag(formula_as_tree, 0, NCDE);
  } else {
    if (nnf_conversion || !negate)
      /* Formula was (eventually) negated by NNF conversion */ 
      dag_ptr = Make_dag(formula_as_tree, 0, NCDE); 
    else
      /* Formula must be negated here */
      dag_ptr = Make_dag(Make_formula_nary(not_code, empty_code, formula_as_tree), 0, NCDE);  
  }
  /* Replacing boxes with modal atoms */
  Replace_boxes(dag_ptr);  
  /* Initializing the decision algorithm */
  main_image_ptr = Init_cnf_sat(dag_ptr, GREEDY_CNF_CONV);  
  Stop_clock(main_monitor_ptr, BUILD);

  /* Eventually, command line parameters change default initialization */
  parse_cmd_line(main_image_ptr, argc, argv);

#ifdef DEBUG
  Print_dag_info(dag_ptr); 
#endif

  /* Some setup */
  Setup_memoizing(dag_ptr);
  Image_monitor_ptr(main_image_ptr) = main_monitor_ptr;
  Test_star_consistency = Choose_test();

  /* Starting the decision algorithm */
  Start_clock(main_monitor_ptr, DSAT); 
  res = Test_dp_sat(main_image_ptr, dag_ptr, Test_star_consistency);
  Stop_clock(main_monitor_ptr, DSAT); 

  /* Get memory allocation data*/
  Inc_param_by(main_monitor_ptr, MAX_SATO_MEMORY, Memory_count(main_image_ptr));

  /* Displaying results */
  Display_results(main_monitor_ptr);
  
  /* Clear things away */
  Clear_image(main_image_ptr);
  Clear_monitor(main_monitor_ptr);

  return (0);
}

Cons_func_ptr_t Choose_test()
{
#ifdef TRANSLATE_TO_K
  return (Test_K_consistency);
#else

#ifdef KSAT
  return (Test_K_consistency);
#endif

#ifdef ESAT
  return (Test_E_consistency);
#endif

#endif
}

void Setup_memoizing(dag_t *dag_ptr)
{
  Force_no_memoT();
  Force_no_memoS();

  if (CACHE_MEMO == 1) {
#ifdef OPTIMIZE
    Init_memoT(Lma(dag_ptr) + 1);
#else
    Init_memoT();
#endif
  } else if (CACHE_MEMO >= 2) {
    Init_memoS(Lma(dag_ptr));
    
    if ((CACHE_MEMO == 3) || (CACHE_MEMO == 5)) Setf_memoS(TRASH_SUBSETS); 
    if ((CACHE_MEMO == 4) || (CACHE_MEMO == 5)) Setf_memoS(GET_DEPENDENCIES); 
  }

  return;
}

void Set_preprocessing_flags(int *bary, int *nnf_conversion, int *negate)
{
#ifdef BARY
  *bary = 1;
#else
  *bary = 0;
#endif
  
#ifdef NNF_CONVERSION
  *nnf_conversion = 1;
#else
  *nnf_conversion = 0;
#endif
  
#ifdef NEGATE
  *negate = 1;
#else
  *negate = 0;
#endif
  
  return;
}


void Translate_to_K(fnode_t *formula_fnode_ptr)
{
#ifdef ESAT
  Translate_from_E_to_K(formula_fnode_ptr);
#endif
}

int Skip_params(int argc, char **argv)
{
  int i;
  
  for(i = 1; i < argc; i++) 
    if (argv[i][0] != '-') break;

  return i;
}

void Help_and_exit(char **argv)
{
  printf("Welcome to *SAT, a platform for experimenting with SAT-based procedures!\n");
  printf("Usage: %s [<Options>] <Input_file>\n\n", argv[0]);
  printf("Defaults are: -s2 -l1 (CHOICE1 = FF, CHOICE2 = TT)\n");
  printf("              -g0, pure literal and relaxation to horn cl. are disabled\n");
  printf("              splitting on dependent variables is disabled\n");
  printf("              early pruning, backjumping, memoizing are disabled\n");
  printf("              -o1000, output is compact\n\n");
  printf("Options are:\n");
  printf("********** SPLITTING **********\n");
  printf("   -s<n>    : choose a different splitting heuristic (n = 1..6):\n");
  printf("               n = 1 MOM strategy to all active variables\n");
  printf("               n = 2 MOM strategy to all active variables in the\n"); 
  printf("                     shortest clauses\n");
  printf("               n = 3 choose the min variable in the first shortest\n");
  printf("                      clause (sign = CHOICE1)\n");
  printf("               n = 4 choose a variable in a shortest clause using\n");
  printf("                     occurrence as the second criterion (sign = CHOICE1)\n");
  printf("               n = 5 choose a variable with the highest occurrence\n");
  printf("                     (sign = CHOICE1)\n");
  printf("               n = 6 choose a variable with the highest occurence\n");
  printf("                     (sign = ?)\n");
  printf("   -l<n>    : choose a different MOM strategy (n = 1..3 - 11..13):\n");
  printf("               n < 11       original SATO weighting:\n");
  printf("                   n=1 sign =(neg_count>pos_count)? CHOICE1 : CHOICE2;\n");
  printf("                   n=2 sign = FF\n");
  printf("                   n=3 sign = TT\n");
  printf("               11 <= n < 21 clauses lenght does not matter (sign as n - 10)\n");
  printf("   -v       : splitting bias becomes CHOICE1 = TT, CHOICE2 = FF\n\n");

  printf("********** OPTIMIZATIONS (PROPOSITIONAL LOGIC ONLY!!)  **********\n");
  printf("   -p       : enable pure literal detection     \n");    
  printf("   -g<n>    : enable propositional backjumping  \n");        
  printf("   -h       : enable relaxation to horn clauses \n\n");        

  printf("********** INTERNAL CNF CONVERSION **********\n");
  printf("   -d       : split on dependent variables\n\n");

  printf("********** MODAL LOGICS **********\n");
  printf("   -e       : enable modal early pruning\n");
  printf("   -b<n>    : enable modal backjumping\n");
  printf("               n=1 try to find smallest inconsistent kernel\n");
  printf("               n=2 discard irrelevant atoms only\n");
  printf("   -m<n>    : enable memoizing:\n");
  printf("               n=1 table memoizing (E, EM, EN, EC, ECN logics only)\n");
  printf("               n=2 set memoizing (K only)\n");
  printf("               n=3 as 2 but also try to be conservative, i.e.,\n");
  printf("                   try to trash useless data first\n");    
  printf("               n=4 as 2 but also store dependencies\n");
  printf("               n=5 as 3 but also store dependencies\n");
  printf("   -w<n>    : cache memory size (sets stored)\n");

  printf("********** MISCELLANEA **********\n");
  printf("   -r       : give a rich output\n");
  printf("   -o<sec>  : enable a timeout (disable any timeout if 0)\n\n");
  
  exit (0);
}

void Display_results(monitor_t *main_monitor_ptr)
{
  if (VERBOSE) {
    printf("The formula was %s\n", (res ? "satisfiable" : "unsatisfiable"));
    printf("--------------------TIMINGS\n");
    printf("Parse time        %s\n", Print_clock(main_monitor_ptr, PARSE));
    printf("Preprocess time   %s\n", Print_clock(main_monitor_ptr, PREPROCESS));
    printf("Translate time    %s\n", Print_clock(main_monitor_ptr, TRANSLATE));
    printf("Build time        %s\n", Print_clock(main_monitor_ptr, BUILD));
    printf("Decision time     %s\n", Print_clock(main_monitor_ptr, DSAT));
    printf("--------------------GENERAL\n");
    printf("Recursive calls   %d\n", Get_param(main_monitor_ptr, REC_CALLS));
    printf("--------------------DAVIS PUTNAM\n");
    printf("Unit propagations %d\n", Get_param(main_monitor_ptr, UNIT_NUM));
    printf("Splits            %d\n", Get_param(main_monitor_ptr, SPLIT_NUM));
    printf("--------------------CNF CONVERSION\n");
    printf("Added variables   %d\n", Laa(dag_ptr) - Lma(dag_ptr));
    if (Use_memoS()) {
      printf("--------------------CACHING\n");
      printf("Cache accesses      %d\n", Stats_memoS(MEMO_ACCESS));
      printf("Cache successes     %d\n", Stats_memoS(MEMO_SUCCESS));
      printf("Items stored        %d (trashing ratio %.2f)\n", 
	     Stats_memoS(MEMO_STORAGE),
	     (float)(Stats_memoS(MEMO_STORAGE) - WINDOW_SIZE) / (float)(Stats_memoS(MEMO_STORAGE))); 
      printf("Dependencies tried  %d\n", Stats_memoS(DEP_ACCESS));
      printf("Dependencies found  %d\n", Stats_memoS(DEP_SUCCESS));
      printf("Dependecies stored  %d (trashing ratio %.2f)\n", 
	     Stats_memoS(DEP_STORAGE),
	     (float)(Stats_memoS(DEP_STORAGE) - WINDOW_SIZE) / (float)(Stats_memoS(DEP_STORAGE))); 
    }
    printf("--------------------MEMORY\n");
    printf("Maximum memory allocated by SATO               %d Kb\n", 
	   (Get_param(main_monitor_ptr, MAX_SATO_MEMORY) /1024));
    if (Use_memoS()) {
      printf("Total memory allocated by CACHE + DEPENDENCIES %d Kb\n", (Stats_memoS(MEMORY) / 1024));
    }
    if (TIMEOUT == -1) printf("TIMEOUT WAS REACHED!!!!\n");
  } else {
    if (TIMEOUT != -1) {
      printf("%s %s %d %d %d %d ",
	     ((TIMEOUT == -1) ? "B" : (res ? "S" : "U")), 
	     Print_clock(main_monitor_ptr, DSAT), 
	     Get_param(main_monitor_ptr, REC_CALLS),
	     Get_param(main_monitor_ptr, UNIT_NUM),
	     Get_param(main_monitor_ptr, SPLIT_NUM),
	     (Laa(dag_ptr) - Lma(dag_ptr)));
      if (Use_memoS()) {
	printf("%d %d %d %.2f %d %d %d %.2f ",
	       Stats_memoS(MEMO_ACCESS),
	       Stats_memoS(MEMO_SUCCESS),
	       Stats_memoS(MEMO_STORAGE),
	       (float)(Stats_memoS(MEMO_STORAGE) - WINDOW_SIZE) / (float)(Stats_memoS(MEMO_STORAGE)),
	       Stats_memoS(DEP_ACCESS),
	       Stats_memoS(DEP_SUCCESS),
	       Stats_memoS(DEP_STORAGE),
	       (float)(Stats_memoS(DEP_STORAGE) - WINDOW_SIZE) / (float)(Stats_memoS(DEP_STORAGE)));
      }
      printf("%d Kb ",
	     (Get_param(main_monitor_ptr, MAX_SATO_MEMORY) / 1024));
      if (Use_memoS()) {
	printf("%d Kb ",
	       (Stats_memoS(MEMORY) / 1024));
      }
      printf("\n");
    }
  }

  return;
}
